from unittest import TestCase
from tgt_grease.enterprise.Detectors import DateDelta
import datetime


class TestDateDelta(TestCase):

    def test_range_compare_pass_lt(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        self.assertTrue(r.timeCompare(
                compare_date,
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '<',
                    'format': '%Y-%m-%d'
                }
            )
        )
        self.assertTrue(r.timeCompare(
                (
                    datetime.datetime.strptime(compare_date, '%Y-%m-%d')
                    + datetime.timedelta(days=1) * -1
                ).strftime('%Y-%m-%d'),
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'past',
                    'operator': '=',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_pass_lt_with_date(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        self.assertTrue(r.timeCompare(
                compare_date,
                {
                    'date': compare_date,
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '<',
                    'format': '%Y-%m-%d'
                }
            )
        )
        self.assertTrue(r.timeCompare(
                (
                    datetime.datetime.strptime(compare_date, '%Y-%m-%d')
                    + datetime.timedelta(days=1) * -1
                ).strftime('%Y-%m-%d'),
                {
                    'date': compare_date,
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'past',
                    'operator': '=',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_pass_lteq(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        self.assertTrue(r.timeCompare(
                compare_date,
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '<=',
                    'format': '%Y-%m-%d'
                }
            )
        )
        self.assertTrue(r.timeCompare(
                (
                    datetime.datetime.strptime(compare_date, '%Y-%m-%d')
                    + datetime.timedelta(days=1) * -1
                ).strftime('%Y-%m-%d'),
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'past',
                    'operator': '<=',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_pass_lteq_with_date(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        self.assertTrue(r.timeCompare(
                compare_date,
                {
                    'date': compare_date,
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '<=',
                    'format': '%Y-%m-%d'
                }
            )
        )
        self.assertTrue(r.timeCompare(
                (
                    datetime.datetime.strptime(compare_date, '%Y-%m-%d')
                    + datetime.timedelta(days=1) * -1
                ).strftime('%Y-%m-%d'),
                {
                    'date': compare_date,
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'past',
                    'operator': '<=',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_pass_gt(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        compare_date = (
                    datetime.datetime.strptime(compare_date, '%Y-%m-%d')
                    + datetime.timedelta(days=2)
                ).strftime('%Y-%m-%d')
        self.assertTrue(r.timeCompare(
                compare_date,
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '>',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_pass_gt_with_date(self):
        r = DateDelta()
        compare_date1 = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        compare_date = (
                    datetime.datetime.strptime(compare_date1, '%Y-%m-%d')
                    + datetime.timedelta(days=2)
                ).strftime('%Y-%m-%d')
        self.assertTrue(r.timeCompare(
                compare_date,
                {
                    'date': compare_date1,
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '>',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_pass_gteq(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        compare_date = (
                    datetime.datetime.strptime(compare_date, '%Y-%m-%d')
                    + datetime.timedelta(days=2)
                ).strftime('%Y-%m-%d')
        self.assertTrue(r.timeCompare(
                compare_date,
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '>=',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_pass_gteq_with_date(self):
        r = DateDelta()
        compare_date1 = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        compare_date = (
                    datetime.datetime.strptime(compare_date1, '%Y-%m-%d')
                    + datetime.timedelta(days=2)
                ).strftime('%Y-%m-%d')
        self.assertTrue(r.timeCompare(
                compare_date,
                {
                    'date': compare_date1,
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '>=',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_pass_neq(self):
        r = DateDelta()
        compare_date1 = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        compare_date = (
                    datetime.datetime.strptime(compare_date1, '%Y-%m-%d')
                    + datetime.timedelta(days=2)
                ).strftime('%Y-%m-%d')
        self.assertTrue(r.timeCompare(
                compare_date,
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '!=',
                    'format': '%Y-%m-%d'
                }
            )
        )
        self.assertTrue(r.timeCompare(
                (
                    datetime.datetime.strptime(compare_date, '%Y-%m-%d')
                    + datetime.timedelta(days=1) * -1
                ).strftime('%Y-%m-%d'),
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'past',
                    'operator': '!=',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_bad_operator(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        self.assertFalse(r.timeCompare(
                compare_date,
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '<>',
                    'format': '%Y-%m-%d'
                }
            )
        )
        self.assertFalse(r.timeCompare(
                (
                    datetime.datetime.strptime(compare_date, '%Y-%m-%d')
                    + datetime.timedelta(days=1) * -1
                ).strftime('%Y-%m-%d'),
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'past',
                    'operator': '<>',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_bad_direction(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        self.assertFalse(r.timeCompare(
                compare_date,
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'cats',
                    'operator': '<',
                    'format': '%Y-%m-%d'
                }
            )
        )
        self.assertFalse(r.timeCompare(
                (
                    datetime.datetime.strptime(compare_date, '%Y-%m-%d')
                    + datetime.timedelta(days=1) * -1
                ).strftime('%Y-%m-%d'),
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'dogs',
                    'operator': '<',
                    'format': '%Y-%m-%d'
                }
            )
        )

    def test_range_compare_value_error(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        self.assertFalse(r.timeCompare(
                compare_date,
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'future',
                    'operator': '<>',
                    'format': 'Y%-%m-%d'
                }
            )
        )
        self.assertFalse(r.timeCompare(
                (
                    datetime.datetime.strptime(compare_date, '%Y-%m-%d')
                    + datetime.timedelta(days=1) * -1
                ).strftime('%Y-%m-%d'),
                {
                    'delta': 'days',
                    'delta_value': 1,
                    'direction': 'past',
                    'operator': '<>',
                    'format': 'Y%-%m-%d'
                }
            )
        )

    def test_good_config(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        result, data = r.processObject(
                {
                    'data': compare_date
                },
                [
                    {
                        'field': 'data',
                        'delta': 'days',
                        'delta_value': 1,
                        'direction': 'future',
                        'operator': '<',
                        'format': '%Y-%m-%d',
                        'variable': True,
                        'variable_name': 'data'
                    }
                ]
            )
        self.assertTrue(result)
        self.assertEqual(data.get('data'), compare_date)
        result, data = r.processObject(
                {
                    'data': compare_date
                },
                [
                    {
                        'field': 'data',
                        'delta': 'days',
                        'delta_value': 1,
                        'direction': 'future',
                        'operator': '<',
                        'format': '%Y-%m-%d'
                    }
                ]
            )
        self.assertTrue(result)
        self.assertFalse(data)

    def test_bad_config(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        result, data = r.processObject(
                {
                    'data': compare_date
                },
                [
                    {
                        'field': 'data',
                        'delta': 'day',
                        'delta_value': 1,
                        'direction': 'future',
                        'operator': '<',
                        'format': '%Y-%m-%d'
                    }
                ]
            )
        self.assertFalse(result)
        self.assertFalse(data)

    def test_field_not_exist(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        result, data = r.processObject(
                {
                },
                [
                    {
                        'field': 'data',
                        'delta': 'day',
                        'delta_value': 1,
                        'direction': 'future',
                        'operator': '<',
                        'format': '%Y-%m-%d'
                    }
                ]
            )
        self.assertFalse(result)
        self.assertFalse(data)

    def test_bad_field(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        result, data = r.processObject(
                {
                    'data': False
                },
                [
                    {
                        'field': 'data',
                        'delta': 'day',
                        'delta_value': 1,
                        'direction': 'future',
                        'operator': '<',
                        'format': '%Y-%m-%d'
                    }
                ]
            )
        self.assertFalse(result)
        self.assertFalse(data)

    def test_pass_out_of_range(self):
        r = DateDelta()
        compare_date = datetime.datetime.utcnow().strftime('%Y-%m-%d')
        result, data = r.processObject(
                {
                    'data': compare_date
                },
                [
                    {
                        'field': 'data',
                        'delta': 'weeks',
                        'delta_value': 2,
                        'direction': 'past',
                        'operator': '<',
                        'format': '%Y-%m-%d',
                        'variable': True,
                        'variable_name': 'data'
                    }
                ]
            )
        self.assertFalse(result)
        self.assertFalse(data)



from tgt_grease.enterprise.Model import Detector
import datetime


class DateDelta(Detector):
    """Date Delta Detector for GREASE Detection

    This detector differs from DateRange as it is relative. In DateRange you can determine constant days whereas
    DateDelta can tell you how many days in the future or past a field is from the date either specified or the
    current date.

    A Typical DateDelta configuration looks like this::

        {
            ...
            'logic': {
                'DateDelta': [
                    {
                        'field': String, # <-- Field to search for
                        'delta': String, # <-- timedelta key for delta range; Accepted Values: weeks, days, hours, minutes, seconds, milliseconds, microseconds,
                        'delta_value': Int, # <-- numeric value for delta to be EX: 1 weeks
                        'format': '%Y-%m-%d', # <-- Mandatory via strptime behavior
                        'operator': String, # <-- Accepted Values: < <= > >= = !=
                        'direction': String, # <-- Accepted Values: future past
                        'date': String, # <-- OPTIONAL, if set then operation will be performed on this date compared to field
                        'variable': Boolean, # <-- OPTIONAL, if true then create a context variable of result
                        'variable_name: String # <-- REQUIRED IF variable, name of context variable
                    }
                    ...
                ]
                ...
            }
        }

    Note:
        Change the format to any supported https://docs.python.org/2/library/datetime.html#strftime-and-strptime-behavior
    Note:
        If date field not provided will be assumed to be UTC Time

    """

    def processObject(self, source, ruleConfig):
        """Processes an object and returns valid rule data

        Data returned in the second parameter from this method should be in this form::

            {
                '<field>': Object # <-- if specified as a variable then return the key->Value pairs
                ...
            }

        Args:
            source (dict): Source Data
            ruleConfig (list[dict]): Rule Configuration Data

        Return:
            tuple: first element boolean for success; second dict for any fields returned as variables

        """
        finalBool = False
        final = {}
        # type checks
        if not isinstance(source, dict):
            return False, {}
        if not isinstance(ruleConfig, list):
            return False, {}
        for block in ruleConfig:
            if not isinstance(block, dict):
                self.ioc.getLogger().error(
                    "INVALID DATERANGE LOGICAL BLOCK! NOT TYPE LIST [{0}]".format(str(type(block))),
                    notify=False
                )
                return False, {}
            # ensure field is there and date format
            if not block.get('field') in source \
                    or 'format' not in block \
                    or not block.get('delta') \
                    or not block.get('operator') \
                    or not block.get('direction') \
                    or block.get('delta') not in ['weeks', 'days', 'hours', 'minutes', 'seconds', 'milliseconds', 'microseconds'] \
                    or not block.get('delta_value'):
                self.ioc.getLogger().error(
                    "malformed rule block; fields [delta, operator, format] are required but not found"
                    " or field not found in source",
                    notify=False
                )
                return False, {}
            if not source.get(block.get('field')):
                self.ioc.getLogger().error(
                    "field equated to False!",
                    notify=False
                )
                return False, {}
            if self.timeCompare(source.get(block.get('field')), block):
                finalBool = True
                if block.get('variable') and block.get('variable_name'):
                    final[str(block.get('variable_name'))] = source.get(block.get('field'))
                else:
                    continue
            else:
                self.ioc.getLogger().trace("Field Failed Range Comparison", verbose=True)
                return False, {}
        return finalBool, final

    def timeCompare(self, field, LogicalBlock):
        """Compares a date to find a delta

        Args:
            field (str): field to compare
            LogicalBlock (dict): Logical Block

        Returns:
            bool: if the range is successful then true else false

        """
        try:
            source_date = datetime.datetime.strptime(field, LogicalBlock.get('format'))
            if LogicalBlock.get('direction') is 'future':
                direction = 1
            elif LogicalBlock.get('direction') is 'past':
                direction = -1
            else:
                self.ioc.getLogger().error("key [direction] is not future or past", notify=False)
                return False
            # setup compare object
            if LogicalBlock.get('date'):
                compare_date = datetime.datetime.strptime(LogicalBlock.get('date'), LogicalBlock.get('format')) \
                               + datetime.timedelta(**{
                                    str(LogicalBlock.get('delta')): int(LogicalBlock.get('delta_value'))
                               }) \
                               * direction
            else:
                compare_date = datetime.datetime.utcnow().strftime(LogicalBlock.get('format'))
                compare_date = datetime.datetime.strptime(compare_date, LogicalBlock.get('format')) \
                               + datetime.timedelta(**{
                                    str(LogicalBlock.get('delta')): int(LogicalBlock.get('delta_value'))
                               }) \
                               * direction
            if LogicalBlock.get('operator') == '<':
                ReturnBool = (source_date < compare_date)
            elif LogicalBlock.get('operator') == '<=':
                ReturnBool = (source_date <= compare_date)
            elif LogicalBlock.get('operator') == '>':
                ReturnBool = (source_date > compare_date)
            elif LogicalBlock.get('operator') == '>=':
                ReturnBool = (source_date >= compare_date)
            elif LogicalBlock.get('operator') == '=':
                ReturnBool = (source_date == compare_date)
            elif LogicalBlock.get('operator') == '!=':
                ReturnBool = (source_date != compare_date)
            else:
                self.ioc.getLogger().error(
                    "Invalid operator provided [{0}]".format(LogicalBlock.get('operator')),
                    notify=False
                )
                return False
            return ReturnBool
        except ValueError:
            # probable datetime format error
            self.ioc.getLogger().error("Value error processing rule!", notify=False)
            return False
        except TypeError:
            # probable datetime format error
            self.ioc.getLogger().error("Type error processing rule!", notify=False)
            return False



from unittest import TestCase
from mock import MagicMock, patch
from tgt_grease.core import GreaseContainer, Configuration
from tgt_grease.enterprise.Model import KafkaSource
import time
import kafka
import multiprocessing as mp

class MockThread():
    def __init__(self):
        self.alive = False
        self.daemon = False
        self.is_alive_called = 0
        self.start_called = 0

    def is_alive(self):
        self.is_alive_called += 1
        return self.alive

    def start(self):
        self.start_called += 1

class TestKafka(TestCase):

    def setUp(self):
        self.ioc = GreaseContainer()
        self.good_config = {"source": "kafka", "max_backlog": 20, "min_backlog": 5, "servers": ["server"], "topics": ["topic"]}
        self.bad_config = {"source": "not kafka"}
        self.mock_thread = MagicMock()
        self.ks = KafkaSource()

    @patch('tgt_grease.enterprise.Model.KafkaSource.validate_configs')
    def test_run_bad_config(self, mock_validate):
        mock_validate.return_value = False
        self.assertFalse(self.ks.run(self.bad_config))

    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_manager_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.validate_configs')
    def test_run_good_config(self, mock_validate, mock_create):
        mock_thread = MockThread()
        mock_create.return_value = mock_thread
        mock_validate.return_value = True
        self.assertFalse(self.ks.run(self.good_config))
        self.assertEqual(self.ks.configs, [self.good_config])
        self.assertEqual(mock_thread.is_alive_called, 1)

    @patch('tgt_grease.enterprise.Model.KafkaSource.get_configs')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_manager_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.validate_configs')
    def test_run_no_config(self, mock_validate, mock_create, mock_get_configs):
        mock_get_configs.return_value = [self.good_config]*5
        mock_thread = MockThread()
        mock_create.return_value = mock_thread
        mock_validate.return_value = True
        self.assertFalse(self.ks.run())
        self.assertEqual(self.ks.configs, [self.good_config]*5)
        self.assertEqual(mock_thread.is_alive_called, 5)

    @patch('tgt_grease.enterprise.Model.KafkaSource.get_configs')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_manager_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.validate_configs')
    def test_run_invalid_config(self, mock_validate, mock_create, mock_get_configs):
        mock_get_configs.return_value = [self.good_config]*5
        mock_thread = MockThread()
        mock_create.return_value = mock_thread
        mock_validate.return_value = False
        self.assertFalse(self.ks.run())
        self.assertEqual(self.ks.configs, [self.good_config]*5)
        self.assertEqual(mock_thread.is_alive_called, 0)

    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.reallocate_consumers')
    def test_consumer_manager(self, mock_reallocate, mock_create, mock_make):
        mock_make.return_value = []
        mock_thread = MockThread()
        mock_create.return_value = (mock_thread, None)
        self.assertFalse(self.ks.consumer_manager(self.ioc, self.good_config))
        self.assertEqual(mock_thread.is_alive_called, 1)
        self.assertEqual(mock_reallocate.call_count, 1)
        self.assertEqual(mock_make.call_count, 1)
        self.assertEqual(mock_create.call_count, 1)

    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer')
    def test_consume_empty_consumer(self, mock_make):
        # It's hard to test something that is designed to run forever, so going to test when the consumer is empty
        mock_make.return_value = []
        pipe1, pipe2 = mp.Pipe()
        self.assertFalse(self.ks.consume(self.ioc, self.good_config, pipe1))

    @patch('tgt_grease.enterprise.Model.KafkaSource.parse_message')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer')
    def test_consume_kill_signal(self, mock_make, mock_parse):
        # It's hard to test something that is designed to run forever, so going to test the pipe kill signal
        mock_make.return_value = ["consumer"]
        pipe1, pipe2 = mp.Pipe()
        pipe1.send("STOP")
        self.assertFalse(self.ks.consume(self.ioc, self.good_config, pipe2))
        mock_parse.assert_not_called()

    def test_sleep(self):
        sleep_time = 1.
        now = time.time()
        self.ks.sleep(sleep_time)
        wake = time.time()
        self.assertTrue(wake-now >= sleep_time)
        self.assertTrue(wake-now < sleep_time + .1)

    def test_parse_message_key_present(self):
        parse_config = {
            "source": "kafka",
            "key_aliases": {"a.b.c": "key"}
        }
        message = MagicMock()
        MagicMock.value = '{"a": {"b": {"c": "value"}}}'
        expected = {"key": "value"}
        self.assertEqual(self.ks.parse_message(self.ioc, parse_config, message), expected)

    def test_parse_message_invalid_json(self):
        parse_config = {
            "source": "kafka",
            "key_aliases": {"a.b.c": "key"}
        }
        message = MagicMock()
        MagicMock.value = '{"a": {"b": {"c": "value"'
        expected = {}
        self.assertEqual(self.ks.parse_message(self.ioc, parse_config, message), expected)

    def test_parse_message_key_present_split(self):
        parse_config = {
            "source": "kafka",
            "key_sep": "@@",
            "key_aliases": {"a@@b@@c": "key"}
        }
        message = MagicMock()
        MagicMock.value = '{"a": {"b": {"c": "value"}}}'

        expected = {"key": "value"}
        self.assertEqual(self.ks.parse_message(self.ioc, parse_config, message), expected)

    def test_parse_message_key_missing(self):
        parse_config = {
            "source": "kafka",
            "key_aliases": {"a.b.c": "key"}
        }
        message = MagicMock()
        MagicMock.value = '{"a": {"b": {"d": "value"}}}'
        expected = {}
        self.assertEqual(self.ks.parse_message(self.ioc, parse_config, message), expected)

    def test_parse_message_key_missing_split(self):
        parse_config = {
            "source": "kafka",
            "split_char": "@@",
            "key_aliases": {"a@@b@@c": "key"}
        }
        message = MagicMock()
        MagicMock.value = '{"a": {"b": {"d": "value"}}}'
        expected = {}
        self.assertEqual(self.ks.parse_message(self.ioc, parse_config, message), expected)

    def test_parse_message_keys_present(self):
        parse_config = {
            "source": "kafka",
            "key_aliases": {"a.b.c": "abc_key",
                            "a.b.d": "abd_key",
                            "aa": "aa_key"
                           }
        }
        message = MagicMock()
        MagicMock.value = '{"a": {"b": {"c": "cvalue", "d":"dvalue"}}, "aa": "aavalue"}'
        expected = {"abc_key": "cvalue", "abd_key": "dvalue", "aa_key": "aavalue"}
        self.assertEqual(self.ks.parse_message(self.ioc, parse_config, message), expected)

    def test_parse_message_keys_missing(self):
        parse_config = {
            "source": "kafka",
            "key_aliases": {"a.b.c": "abc_key",
                            "a.b.d": "abd_key",
                            "aa": "aa_key"
                           }
        }
        message = MagicMock()
        MagicMock.value = '{"a": {"b": {"c": "cvalue"}}, "aa": "aavalue"}'
        expected = {}
        self.assertEqual(self.ks.parse_message(self.ioc, parse_config, message), expected)

    @patch('tgt_grease.enterprise.Model.KafkaSource.sleep')
    @patch('tgt_grease.enterprise.Model.KafkaSource.kill_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.get_backlog')
    def test_reallocate_consumers_kill(self, mock_backlog, mock_create, mock_kill, mock_sleep):
        mock_backlog.return_value = 3
        self.assertEqual(self.ks.reallocate_consumers(self.ioc, self.good_config, None, ["thread1", "thread2"]), -1)
        mock_kill.assert_called_once_with(self.ioc, "thread1")
        self.assertEqual(mock_backlog.call_count, 2)
        mock_create.assert_not_called()

    @patch('tgt_grease.enterprise.Model.KafkaSource.sleep')
    @patch('tgt_grease.enterprise.Model.KafkaSource.kill_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.get_backlog')
    def test_reallocate_consumers_kill_1thread(self, mock_backlog, mock_create, mock_kill, mock_sleep):
        mock_backlog.return_value = 3
        self.assertEqual(self.ks.reallocate_consumers(self.ioc, self.good_config, None, ["thread1"]), 0)
        mock_kill.assert_not_called()
        self.assertEqual(mock_backlog.call_count, 2)
        mock_create.assert_not_called()

    @patch('tgt_grease.enterprise.Model.KafkaSource.sleep')
    @patch('tgt_grease.enterprise.Model.KafkaSource.kill_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.get_backlog')
    def test_reallocate_consumers_create(self, mock_backlog, mock_create, mock_kill, mock_sleep):
        mock_backlog.return_value = 21
        mock_create.return_value = "new_thread"
        threads = ["thread1"]
        self.assertEqual(self.ks.reallocate_consumers(self.ioc, self.good_config, None, threads), 1)
        mock_kill.assert_not_called()
        self.assertEqual(mock_backlog.call_count, 2)
        mock_create.assert_called_once_with(self.ioc, self.good_config)
        self.assertTrue("new_thread" in threads)

    @patch('tgt_grease.enterprise.Model.KafkaSource.sleep')
    @patch('tgt_grease.enterprise.Model.KafkaSource.kill_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.get_backlog')
    def test_reallocate_consumers_pass(self, mock_backlog, mock_create, mock_kill, mock_sleep):
        mock_backlog.return_value = 10
        self.assertEqual(self.ks.reallocate_consumers(self.ioc, self.good_config, None, ["thread1"]), 0)
        mock_kill.assert_not_called()
        mock_create.assert_not_called()
        self.assertEqual(mock_backlog.call_count, 2)


    @patch('tgt_grease.enterprise.Model.KafkaSource.sleep')
    @patch('tgt_grease.enterprise.Model.KafkaSource.kill_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.get_backlog')
    def test_reallocate_consumers_max_thread(self, mock_backlog, mock_create, mock_kill, mock_sleep):
        mock_backlog.return_value = 30
        threads = ["thread" for i in range(0, 32)]
        self.assertEqual(self.ks.reallocate_consumers(self.ioc, self.good_config, None, threads), 0)
        mock_kill.assert_not_called()
        mock_create.assert_not_called()
        self.assertEqual(mock_backlog.call_count, 2)

    @patch('tgt_grease.enterprise.Model.KafkaSource.sleep')
    @patch('tgt_grease.enterprise.Model.KafkaSource.kill_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.get_backlog')
    def test_reallocate_consumers_max_thread_create(self, mock_backlog, mock_create, mock_kill, mock_sleep):
        mock_backlog.return_value = 30
        threads = ["thread" for i in range(0, 33)]
        config = {"max_consumers": 35, "max_backlog":20, "min_backlog":5}
        self.assertEqual(self.ks.reallocate_consumers(self.ioc, config, None, threads), 1)
        mock_kill.assert_not_called()
        mock_create.assert_called()
        self.assertEqual(mock_backlog.call_count, 2)

    @patch('tgt_grease.enterprise.Model.KafkaSource.sleep')
    @patch('tgt_grease.enterprise.Model.KafkaSource.kill_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.create_consumer_thread')
    @patch('tgt_grease.enterprise.Model.KafkaSource.get_backlog')
    def test_reallocate_consumers_max_thread_pass(self, mock_backlog, mock_create, mock_kill, mock_sleep):
        mock_backlog.return_value = 30
        threads = ["thread" for i in range(0, 36)]
        config = {"max_consumers": 35, "max_backlog":20, "min_backlog":5}
        self.assertEqual(self.ks.reallocate_consumers(self.ioc, config, None, threads), 0)
        mock_kill.assert_not_called()
        mock_create.assert_not_called()
        self.assertEqual(mock_backlog.call_count, 2)

    @patch('tgt_grease.enterprise.Model.KafkaSource.sleep')
    def test_kill_consumer_thread(self, mock_sleep):
        conn1, conn2 = mp.Pipe()
        self.ks.kill_consumer_thread(self.ioc, (None, conn1))
        self.assertEqual(conn2.recv(), "STOP")

    def test_get_backlog_happy(self):
        mock_consumer = MagicMock()
        for part_count in range(1, 10):
            for start in range(0, 10):
                for end in range(start, 10):
                    mock_partitions = ["part" + str(part_i) for part_i in range(part_count)] # assignment returns an array of TopicPartitions, but our mocked consumer works with just strings
                    mock_consumer.assignment.return_value = mock_partitions
                    mock_consumer.position.return_value = start
                    mock_consumer.end_offsets.return_value = {part:end for part in mock_partitions}
                    expected_average = end - start
                    res = self.ks.get_backlog(self.ioc, mock_consumer)
                    self.assertTrue(isinstance(res, float))
                    self.assertEqual(res, expected_average)

    def test_get_backlog_not_assigned(self):
        mock_consumer = MagicMock()
        assigned = {"status": False}

        def poll():
            assigned["status"] = True

        mock_consumer.poll.side_effect = poll
        for part_count in range(1, 10):
            for start in range(0, 10):
                for end in range(start, 10):
                    mock_partitions = ["part" + str(part_i) for part_i in range(part_count)] # assignment returns an array of TopicPartitions, but our mocked consumer works with just strings
                    def assignment():
                        if assigned["status"]:
                            return mock_partitions
                        else:
                            return []

                    mock_consumer.assignment.side_effect = assignment
                    mock_consumer.position.return_value = start
                    mock_consumer.end_offsets.return_value = {part:end for part in mock_partitions}
                    expected_average = end - start
                    res = self.ks.get_backlog(self.ioc, mock_consumer)
                    self.assertTrue(isinstance(res, float))
                    self.assertEqual(res, expected_average)

    def test_get_backlog_no_assign(self):
        mock_consumer = MagicMock()
        mock_consumer.assignment.return_value = []
        self.assertEqual(self.ks.get_backlog(self.ioc, mock_consumer), -1.)

    def test_get_backlog_position_error(self):
        mock_consumer = MagicMock()
        mock_consumer.assignment.return_value = ["part"]
        mock_consumer.position.side_effect = kafka.errors.KafkaTimeoutError()
        mock_consumer.end_offsets.return_value = {"part": 1}
        self.assertEqual(self.ks.get_backlog(self.ioc, mock_consumer), -1.)

    def test_get_backlog_end_offsets_error(self):
        mock_consumer = MagicMock()
        mock_consumer.assignment.return_value = ["part"]
        mock_consumer.position.return_value = 1
        mock_consumer.end_offsets.side_effect  = kafka.errors.UnsupportedVersionError()
        self.assertEqual(self.ks.get_backlog(self.ioc, mock_consumer), -1.)

    @patch("tgt_grease.enterprise.Model.CentralScheduling.Scheduling.scheduleDetection")
    def test_send_to_scheduling_happy(self, mock_scheduling):
        config = {
            "source": "kafka",
            "name": "test_config"
        }
        mock_msg = {"a": "b"}
        mock_scheduling.return_value = True
       
        self.assertTrue(self.ks.send_to_scheduling(self.ioc, config, mock_msg))
        mock_scheduling.assert_called_once_with("kafka", "test_config", mock_msg)

    @patch("tgt_grease.enterprise.Model.CentralScheduling.Scheduling.scheduleDetection")
    def test_send_to_scheduling_sad(self, mock_scheduling):
        config = {
            "source": "kafka",
            "name": "test_config"
        }
        mock_msg = {"a": "b"}
        mock_scheduling.return_value = False
        self.assertFalse(self.ks.send_to_scheduling(self.ioc, config, mock_msg))
        mock_scheduling.assert_called_once_with("kafka", "test_config", mock_msg)

    @patch("threading.Thread")
    def test_create_consumer_manager_thread(self, mock_thread):
        mockp = MockThread()
        mock_thread.return_value = mockp
        self.assertEqual(self.ks.create_consumer_manager_thread(self.good_config), mockp)
        self.assertEqual(mockp.is_alive_called, 0)
        self.assertEqual(mockp.start_called, 1)
        self.assertFalse(mockp.daemon)

    @patch("threading.Thread")
    def test_create_consumer_thread(self, mock_thread):
        mockp = MockThread()
        mock_thread.return_value = mockp
        thread, pipe = self.ks.create_consumer_thread(self.ioc, self.good_config)
        self.assertEqual(thread, mockp)
        self.assertEqual(type(pipe), type(mp.Pipe()[0]))
        self.assertEqual(mockp.is_alive_called, 0)
        self.assertEqual(mockp.start_called, 1)
        self.assertTrue(mockp.daemon)

    def test_validate_configs_happy(self):
        good_config = {
            "name": "kafka_config",
            "source": "kafka",
            "key_aliases": {
                "a*b*c": "abc_key",
                "a*b*d": "abd_key"
            },
            "key_sep": "*",         #opt
            "max_consumers": 32,    #opt
            "topics": [
                "topic1",
                "topic2"
            ],
            "servers": [
                "server.target.com:1234"
            ],
            "max_backlog": 200,     #opt
            "min_backlog": 100      #opt
        }

        self.assertTrue(self.ks.validate_configs([good_config]))
        self.assertTrue(self.ks.validate_configs([good_config]*5))

    def test_validate_configs_wrong_source(self):
        config = {
            "name": "kafka_config",
            "source": "not kafka",
            "key_aliases": {
                "a*b*c": "abc_key",
                "a*b*d": "abd_key"
            },
            "key_sep": "*",         #opt
            "max_consumers": 32,    #opt
            "topics": [
                "topic1",
                "topic2"
            ],
            "servers": [
                "server.target.com:1234"
            ],
            "max_backlog": 200,     #opt
            "min_backlog": 100      #opt
        }

        self.assertFalse(self.ks.validate_configs([config]))
        self.assertFalse(self.ks.validate_configs([config]*5))

    def test_validate_configs_duplicate_aliases(self):
        config = {
            "name": "kafka_config",
            "source": "kafka",
            "key_aliases": {
                "a*b*c": "key",
                "a*b*d": "key"
            },
            "key_sep": "*",         #opt
            "max_consumers": 32,    #opt
            "topics": [
                "topic1",
                "topic2"
            ],
            "servers": [
                "server.target.com:1234"
            ],
            "max_backlog": 200,     #opt
            "min_backlog": 100      #opt
        }

        self.assertFalse(self.ks.validate_configs([config]))
        self.assertFalse(self.ks.validate_configs([config]*5))

    def test_validate_configs_wrong_types(self):
        config = {
            "name": "kafka_config",
            "source": "kafka",
            "key_aliases": {
                "a*b*c": "abc_key",
                "a*b*d": "abd_key"
            },
            "key_sep": 11,         #opt
            "max_consumers": "32", #opt
            "topics": [
                "topic1",
                "topic2"
            ],
            "servers": [
                "server.target.com:1234"
            ],
            "max_backlog": 200,     #opt
            "min_backlog": 100      #opt
        }

        self.assertFalse(self.ks.validate_configs([config]))
        self.assertFalse(self.ks.validate_configs([config]*5))

    def test_validate_configs_key_missing(self):
        config = {
            "name": "kafka_config",
            "source": "kafka",
            "key_sep": "*",         #opt
            "max_consumers": 32,    #opt
            "topics": [
                "topic1",
                "topic2"
            ],
            "servers": [
                "server.target.com:1234"
            ],
            "max_backlog": 200,     #opt
            "min_backlog": 100      #opt
        }

        self.assertFalse(self.ks.validate_configs([config]))
        self.assertFalse(self.ks.validate_configs([config]*5))

    def test_validate_configs_empty_keys(self):
        config = {
            "name": "kafka_config",
            "source": "kafka",
            "key_aliases": {},
            "key_sep": "*",         #opt
            "max_consumers": 32,    #opt
            "topics": [],
            "servers": [],
            "max_backlog": 200,     #opt
            "min_backlog": 100      #opt
        }

        self.assertFalse(self.ks.validate_configs([config]))
        self.assertFalse(self.ks.validate_configs([config]*5))

    def test_validate_configs_no_opt(self):
        config = {
            "name": "kafka_config",
            "source": "kafka",
            "key_aliases": {
                "a*b*c": "abc_key",
                "a*b*d": "abd_key"
            },
            "topics": [
                "topic1",
                "topic2"
            ],
            "servers": [
                "server.target.com:1234"
            ]
        }

        self.assertTrue(self.ks.validate_configs([config]))
        self.assertTrue(self.ks.validate_configs([config]*5))



from tgt_grease.core.Types import Command
from tgt_grease.enterprise.Model import Scan


class Scanner(Command):
    """The Scan Command

    This class is the ingestion of information for GREASE. It utilizes configurations to 'wire' scanners together

    """

    __author__ = "James E. Bell Jr."
    __version__ = "2.0.0"
    purpose = "Parse the configured environment for data and schedule de-duplicated data for detection"
    help = """
    This command scans the environment via the node configuration. This enables GREASE to 'see' its' environment and 
    schedule jobs a detection node
    
    Args:
        --loop:<int>
            How many scans you would like to go through
        --config:<filename>
            The specific config file you want to parse
        --source:<source>
            The specific source to parse all configs of
        --foreground
            Print Log messages to foreground
    """

    def __init__(self):
        super(Scanner, self).__init__()

    def execute(self, context):
        """Execute method of the scanner prototype

        Args:
            context (dict): Command Context

        Note:
            This method normally will *never* return. As it is a prototype. So it should continue into infinity

        Returns:
             bool: True always unless failures occur

        """
        if context.get('foreground'):
            # set foreground if in context
            self.ioc.getLogger().foreground = True
        self.ioc.getLogger().trace("Scanner starting", trace=True)
        scanner = Scan(self.ioc)
        # create Parse Args
        args = {
            'source': context.get('source', self.ioc.getConfig().get('Sourcing', 'source', None)),
            'config': context.get('config', self.ioc.getConfig().get('Sourcing', 'config', None))
        }
        if 'loop' in context:
            # scan only a certain amount of times
            scan_count = 0
            while scan_count < int(context.get('loop')):
                scanner.Parse(**args)
                scan_count += 1
        else:
            try:
                while True:
                    scanner.Parse(**args)
            except KeyboardInterrupt:
                # graceful close for scanning
                self.ioc.getLogger().trace("Keyboard interrupt in scanner detected", trace=True)
                return True
        # ensure we clean up after ourselves
        if context.get('foreground'):
            self.ioc.getLogger().foreground = False
        return True



from logging import DEBUG, ERROR, INFO
from tgt_grease.core import GreaseContainer, ImportTool
from tgt_grease.core.Types import Command
from tgt_grease.enterprise.Model import PrototypeConfig
from datetime import datetime
from bson.objectid import ObjectId
import threading
from psutil import cpu_percent, virtual_memory


class DaemonProcess(object):
    """Actual daemon processing for GREASE Daemon

    Attributes:
        ioc (GreaseContainer): The Grease IOC
        current_real_second (int): Current second in time
        registered (bool): If the node is registered with MongoDB
        impTool (ImportTool): Instance of Import Tool
        conf (PrototypeConfig): Prototype Configuration Instance

    """

    ioc = None
    current_real_second = None
    registered = True
    contextManager = {'jobs': {}, 'prototypes': {}}
    impTool = None

    def __init__(self, ioc):
        if isinstance(ioc, GreaseContainer):
            self.ioc = ioc
        else:
            self.ioc = GreaseContainer()
        self.current_real_second = datetime.utcnow().second
        if self.ioc.getConfig().NodeIdentity == "Unknown" and not self.register():
            self.registered = False
        self.impTool = ImportTool(self.ioc.getLogger())
        self.conf = PrototypeConfig(self.ioc)

    def server(self):
        """Server process for ensuring prototypes & jobs are running

        By Running this method this will clear the DB of any jobs a node may have

        Returns:
            bool: Server Success

        """
        # Ensure we aren't swamping the system
        cpu = cpu_percent(interval=.1)
        mem = virtual_memory().percent
        if \
                cpu >= int(self.ioc.getConfig().get('NodeInformation', 'ResourceMax')) \
                or mem >= int(self.ioc.getConfig().get('NodeInformation', 'ResourceMax')):
            self.ioc.getLogger().trace(
                "Thread Maximum Reached CPU: [{0}] Memory: [{1}]".format(cpu, mem),
                trace=True
            )
            # remove variables
            del cpu
            del mem
            return True
        if not self.registered:
            self.ioc.getLogger().trace("Server is not registered", trace=True)
            return False
        self.ioc.getLogger().trace("Server execution starting", trace=True)
        # establish job collection
        JobsCollection = self.ioc.getCollection("SourceData")
        self.ioc.getLogger().trace("Searching for Jobs", trace=True)
        jobs = JobsCollection.find({
            'grease_data.execution.server': ObjectId(self.ioc.getConfig().NodeIdentity),
            'grease_data.execution.commandSuccess': False,
            'grease_data.execution.executionSuccess': False,
            'grease_data.execution.failures': {'$lt': 6}
        })
        # Get Node Information
        Node = self.ioc.getCollection('JobServer').find_one({'_id': ObjectId(self.ioc.getConfig().NodeIdentity)})
        if not Node:
            # If for some reason we couldn't find it
            self.ioc.getLogger().error("Failed To Load Node Information")
            return False
        # Get Prototypes
        prototypes = list(Node.get('prototypes'))
        # Del node instance
        del Node
        if prototypes:
            # We have prototypes to spin up
            for prototype in prototypes:
                self.ioc.getLogger().trace("Passing ProtoType [{0}] to Runner".format(prototype), trace=True)
                self._run_prototype(prototype)
        if jobs.count():
            self.ioc.getLogger().trace("Total Jobs to Execute: [{0}]".format(jobs.count()))
            for job in jobs:
                self.ioc.getLogger().trace("Passing Job [{0}] to Runner".format(job.get("_id")), trace=True)
                self._run_job(job, JobsCollection)
        else:
            # Nothing to Run for Jobs
            self.ioc.getLogger().trace("No Jobs Scheduled to Server", trace=True)
        self.ioc.getLogger().trace("Server execution complete", trace=True)
        return True

    def _run_job(self, job, JobCollection):
        """Run a On-Demand Job

        Args:
            job (dict): Job Data to execute
            JobCollection (pymongo.collection.Collection): JobCollection to update for telemetry

        Returns:
            None: Void Method to kickoff execution

        """
        if not self.contextManager['jobs'].get(job.get('_id')):
            # New Job to run
            if isinstance(job.get('configuration'), bytes):
                conf = job.get('configuration').decode()
            else:
                conf = job.get('configuration')
            inst = self.impTool.load(self.conf.get_config(conf).get('job'))
            if inst and isinstance(inst, Command):
                inst.ioc.getLogger().foreground = self.ioc.getLogger().foreground
                thread = threading.Thread(
                    target=inst.safe_execute,
                    args=(job.get('grease_data', {}).get('detection', {}).get('detection', {}),),
                    name="GREASE DAEMON COMMAND EXECUTION [{0}]".format(job.get('_id'))
                )
                thread.daemon = True
                thread.start()
                self.contextManager['jobs'][job.get("_id")] = {
                    'thread': thread,
                    'command': inst
                }
            else:
                # Invalid Job
                del inst
                self.ioc.getLogger().warning("Invalid Job", additional=job)
                JobCollection.update_one(
                    {'_id': ObjectId(job['_id'])},
                    {
                        '$set': {
                            'grease_data.execution.failures': job.get('failures', 0) + 1
                        }
                    }
                )
            return
        else:
            # Job already executing
            if self.contextManager['jobs'].get(job.get('_id')).get('thread').isAlive():
                # thread still executing
                return
            else:
                # Execution has ended
                self.ioc.getLogger().trace("Job [{0}] finished running".format(job.get('_id')), trace=True)
                finishedJob = self.contextManager['jobs'].get(job.get('_id')).get('command')  # type: Command
                if finishedJob.getRetVal():
                    # job completed successfully
                    JobCollection.update_one(
                        {'_id': ObjectId(job.get('_id'))},
                        {
                            '$set': {
                                'grease_data.execution.commandSuccess': finishedJob.getRetVal(),
                                'grease_data.execution.executionSuccess': finishedJob.getExecVal(),
                                'grease_data.execution.completeTime': datetime.utcnow(),
                                'grease_data.execution.returnData': finishedJob.getData()
                            }
                        }
                    )
                else:
                    # Job Failure
                    self.ioc.getLogger().warning(
                        "Job Failed [{0}]".format(job.get('_id')), additional=finishedJob.getData()
                    )
                    # TODO: Job Execution cooldown timing
                    JobCollection.update_one(
                        {'_id': ObjectId(job['_id'])},
                        {
                            '$set': {
                                'grease_data.execution.failures': job.get('grease_data', {}).get('execution', {}).get('failures', 0) + 1
                            }
                        }
                    )
                # close out job
                finishedJob.__del__()
                del finishedJob
                # remove from contextManager
                del self.contextManager['jobs'][job.get('_id')]
                return

    def _run_prototype(self, prototype):
        """Startup a ProtoType

        Args:
            prototype (str): ProtoType to start

        Returns:
            None: Void method to start prototype

        """
        if not self.contextManager['prototypes'].get(prototype):
            # ProtoType has not started
            inst = self.impTool.load(prototype)
            if not isinstance(inst, Command):
                # invalid ProtoType
                self.log_once_per_second("Invalid ProtoType [{0}]".format(prototype), level=ERROR)
                return
            inst.ioc.getLogger().foreground = self.ioc.getLogger().foreground
            thread = threading.Thread(
                target=inst.safe_execute,
                args=({}),
                name="GREASE DAEMON PROTOTYPE [{0}]".format(prototype)
            )
            thread.daemon = True
            thread.start()
            self.contextManager['prototypes'][prototype] = thread
            return
        else:
            # ensure thread is alive
            if self.contextManager['prototypes'].get(prototype).isAlive():
                self.ioc.getLogger().trace("ProtoType [{0}] is alive".format(prototype))
                return
            else:
                # Thread died for some reason
                self.log_once_per_second("ProtoType [{0}] Stopped".format(prototype), level=INFO)
                inst = self.impTool.load(prototype)
                if not isinstance(inst, Command):
                    self.log_once_per_second("Invalid ProtoType [{0}]".format(prototype), level=ERROR)
                    return
                inst.ioc.getLogger().foreground = self.ioc.getLogger().foreground
                thread = threading.Thread(
                    target=inst.execute,
                    name="GREASE DAEMON PROTOTYPE [{0}]".format(prototype)
                )
                thread.daemon = True
                thread.start()
                self.contextManager['prototypes'][prototype] = thread
                return

    def drain_jobs(self, JobCollection):
        """Will drain jobs from the current context

        This method is used to prevent abnormal ending of executions

        Args:
            JobCollection (pymongo.collection.Collection): Job Collection Object

        Returns:
            bool: When job queue is emptied

        """
        Threads = True
        while Threads:
            if self.contextManager['jobs']:
                jobs = {}
                for key, val in self.contextManager['jobs'].items():
                    if val['thread'].isAlive():
                        jobs[key] = val
                        continue
                    else:
                        # Execution has ended
                        self.ioc.getLogger().trace("Job [{0}] finished running".format(key), trace=True)
                        finishedJob = self.contextManager['jobs'].get(key).get('command')  # type: Command
                        if finishedJob.getRetVal():
                            # job completed successfully
                            JobCollection.update_one(
                                {'_id': ObjectId(key)},
                                {
                                    '$set': {
                                        'grease_data.execution.commandSuccess': finishedJob.getRetVal(),
                                        'grease_data.execution.executionSuccess': finishedJob.getExecVal(),
                                        'grease_data.execution.completeTime': datetime.utcnow(),
                                        'grease_data.execution.returnData': finishedJob.getData()
                                    }
                                }
                            )
                        else:
                            # Job Failure
                            self.ioc.getLogger().warning(
                                "Job Failed [{0}]".format(key), additional=finishedJob.getData()
                            )
                            JobCollection.update_one(
                                {'_id': ObjectId(key)},
                                {
                                    '$set': {
                                        'grease_data.execution.failures': val['command'].get('failures', 0) + 1
                                    }
                                }
                            )
                        # close out job
                        finishedJob.__del__()
                        del finishedJob
                self.contextManager['jobs'] = jobs
            else:
                Threads = False
        return True

    def register(self):
        """Attempt to register with MongoDB

        Returns:
            bool: Registration Success

        """
        return self.ioc.ensureRegistration()

    def log_once_per_second(self, message, level=DEBUG, additional=None):
        """Log Message once per second

        Args:
            message (str): Message to log
            level (int): Log Level
            additional (object): Additional information that is able to be str'd

        Returns:
            None: Void Method to fire log message

        """
        if self._has_time_progressed():
            self.ioc.getLogger().TriageMessage(message=message, level=level, additional=additional)

    def _has_time_progressed(self):
        """Determines if the current second and the real second are not the same

        Returns:
            bool: if true then time has passed in a meaningful way

        """
        if self.current_real_second != datetime.utcnow().second:
            self.current_real_second = datetime.utcnow().second
            return True
        else:
            return False



from setuptools import setup, find_packages
import os


setup(
    name='tgt_grease',
    version='2.3.2',
    license="MIT",
    description='Modern distributed automation engine built with love by Target',
    long_description="""
    GREASE is a general purpose distributed automation engine designed to scale to enterprise workloads. We utilize
    MongoDB and a plugin architecture to enable broad automation possibilities via one common core of primitives and 
    services. 
    """,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',
        'Topic :: System',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: Microsoft :: Windows :: Windows 8',
        'Operating System :: Microsoft :: Windows :: Windows 8.1',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: POSIX'
    ],
    keywords='python automated recovery',
    author='James E. Bell Jr.',
    author_email="james.e.bell@target.com",
    packages=find_packages(),
    test_suite='nose.collector',
    tests_require=['nose', 'nose-cover3', 'mock'],
    install_requires=[
        'psycopg2',
        'requests',
        'pymongo',
        'psutil',
        'elasticsearch',
        'kafka'
    ] + (
         ["pywin32"] if "nt" == os.name else []
        ),
    include_package_data=True,
    zip_safe=False,
    scripts=[
        'bin/grease',
        'bin/grease.ps1'
    ]
)
